/**
  * JASS Project. WS-CAF.
  *
  * Copyright (c) 2006, Distributed Systems Lab, Universidad Politecnica 
  * de Madrid (Spain)
  * All rights reserved.
  * Contact:
  *        web page: http://lsd.ls.fi.upm.es/lsd
  *        e-mail: mpatino@fi.upm.es or rjimenez@fi.upm.es
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met:
  *
  *    - Redistributions of source code must retain the above copyright 
  *      notice, this list of conditions and the following disclaimer.
  *    - Redistributions in binary form must reproduce the above copyright 
  *      notice, this list of conditions and the following disclaimer in the 
  *      documentation and/or other materials provided with the distribution.
  *    - Neither the name of the Distributed Systems Lab nor the names of 
  *      its contributors may be used to endorse or promote products derived from 
  *      this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, *
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * 
  * Contributors: 
  * Francisco Pérez-Sorrosal (fpsorrosal@zipi.fi.upm.es)
  * Rubén Naranjo Izquierdo (rnaranjo@alumnos.upm.es)
  * José Antonio Sánchez Murillo (jsanchez@alumnos.upm.es)
  * 
  **/

package org.oasis.wscaf.core.wscf.services;

import javax.xml.namespace.QName;

import org.apache.axis2.addressing.EndpointReference;
import org.oasis.wscaf.core.wscf.exceptions.DuplicateParticipant;
import org.oasis.wscaf.core.wscf.exceptions.InvalidProtocol;
import org.oasis.wscaf.core.wscf.exceptions.ParticipantNotFound;
import org.oasis.wscaf.core.wscf.exceptions.ResponseFault;
import org.oasis.wscaf.core.wscf.types.ParticipantGroup;
import org.oasis.wscaf.core.wscf.types.ParticipantPairGroup;
import org.oasis.wscaf.core.wscf.types.RegistrationContext;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidContext;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidState;
import org.oasis.wscaf.core.wsctx.exceptions.NoContext;
import org.oasis.wscaf.core.wsctx.exceptions.UnknownContext;
import org.oasis.wscaf.core.wsctx.services.ContextManager;
import org.oasis.wscaf.core.wsctx.types.CompletionHandler;
import org.oasis.wscaf.core.wsctx.types.ContextIdentifier;
import org.oasis.wscaf.core.wsctx.types.ContextWrapper;

public class RegistrationService {

	// The singleton instance
    private static RegistrationService singleton = null;

	// The context manager
    private ContextManager contextManager = null;

	private static String[] protocolsSupported = {"LRA", "ACID", "BP"};
	
    /**
     * Private constructor.
     */
    private RegistrationService() {
        // Get the context manager
        contextManager = ContextManager.getInstance();
        System.out.println("\t[RegistrationService] WSCAF CTX Manager instance obtained!!!");
    }

    /**
     * Returns the singleton instance of the registration service.
     * 
     * @return The singleton instance.
     */
    public static synchronized RegistrationService getInstance() {
    	System.out.println("[RegistrationService] Returning instance");
        if (singleton == null)
            singleton = new RegistrationService();
        return singleton;
    }
    
    public RegistrationContext addParticipant(ContextIdentifier context, ParticipantGroup[] participantGroups)
    throws ResponseFault { // UnknownContext, InvalidState, DuplicateParticipant, InvalidProtocol
        RegistrationContext newContext=null;
        for (int i=0; i<participantGroups.length; i++)
            newContext = addParticipant(context, participantGroups[i]);
        return newContext;
    }

    public RegistrationContext addParticipant(ContextIdentifier context, ParticipantGroup participants)
		throws ResponseFault { // UnknownContext, InvalidState, DuplicateParticipant, InvalidProtocol

    	ContextWrapper localContext;
    	String[] protocols = participants.getProtocols();
    	try {
			localContext = contextManager.getContext(context);
			checkProtocolsSupported(protocols);
		} catch (InvalidProtocol e) {
			throw new ResponseFault(0, e);
		} catch (NoContext e) {
			throw new ResponseFault(0, new UnknownContext(e.getMessage()));
		} catch (InvalidContext e) {
			throw new ResponseFault(0, new InvalidState(e.getMessage()));
		}
		RegistrationContext registrationContext = RegistrationContext.augmentToRegistrationContext(localContext.getContext());
        
		for (int i=0; i<participants.getParticipantsSize(); i++) {
            try {
                registrationContext.addParticipant(participants.getParticipant(i), protocols);
            } catch (DuplicateParticipant e) {
                throw new ResponseFault(i, new DuplicateParticipant(e.getMessage(), e.getDuplicateParticipant()));
            }
        }
		localContext.setContext(registrationContext);
        addCompletionHandlers(localContext, protocols);

		return registrationContext;
	}

    public RegistrationContext removeParticipant(ContextIdentifier context, ParticipantGroup[] participantGroups)
    throws ResponseFault { // UnknownContext, InvalidState, DuplicateParticipant, InvalidProtocol
        RegistrationContext newContext=null;
        for (int i=0; i<participantGroups.length; i++)
            newContext = removeParticipant(context, participantGroups[i]);
        return newContext;
    }

	public RegistrationContext removeParticipant(ContextIdentifier context, ParticipantGroup participants) 
		throws ResponseFault { // UnknownContext, InvalidState, ParticipantNotFound

		ContextWrapper localContext;
    	try {
			localContext = contextManager.getContext(context);
		} catch (NoContext e) {
			throw new ResponseFault(0, new UnknownContext(e.getMessage()));
		} catch (InvalidContext e) {
			throw new ResponseFault(0, new InvalidState(e.getMessage()));
		}
		RegistrationContext registrationContext = RegistrationContext.augmentToRegistrationContext(localContext.getContext());
		
		String[] protocols = participants.getProtocols();
		for (int i=0;i<protocols.length;i++) {
			for (int j=0; j<participants.getParticipantsSize(); j++) {
				try {
					registrationContext.removeParticipant(protocols[i],participants.getParticipant(j));
				} catch (ParticipantNotFound e) {
					throw new ResponseFault(j, e);
				}
			}
		}
		
		localContext.setContext(registrationContext);
		return registrationContext;
	}

    public RegistrationContext replaceParticipant(ContextIdentifier context, ParticipantPairGroup[] participantPairGroups)
    throws ResponseFault { // UnknownContext, InvalidState, DuplicateParticipant, InvalidProtocol
        RegistrationContext newContext=null;
        for (int i=0; i<participantPairGroups.length; i++)
            newContext = replaceParticipant(context, participantPairGroups[i]);
        return newContext;
    }

	public RegistrationContext replaceParticipant(ContextIdentifier context, ParticipantPairGroup participantPairs)
		throws ResponseFault { // UnknownContext, InvalidState, TransientFault 

		ContextWrapper localContext;
    	try {
			localContext = contextManager.getContext(context);
		} catch (NoContext e) {
			throw new ResponseFault(0, new UnknownContext(e.getMessage()));
		} catch (InvalidContext e) {
			throw new ResponseFault(0, new InvalidState(e.getMessage()));
		}
		RegistrationContext registrationContext = RegistrationContext.augmentToRegistrationContext(localContext.getContext());
		
		String[] protocols = participantPairs.getProtocols();
		for (int j=0;j<protocols.length;j++) {
			for (int i=0; i<participantPairs.getParticipantPairsSize(); i++) { 
				registrationContext.replaceParticipant(protocols[j],participantPairs.getParticipantPair(i));
			}
		}
		
		localContext.setContext(registrationContext);
		return registrationContext;
	}

	public ParticipantGroup getParticipants(ContextIdentifier context, String[] subProtocols)
		throws InvalidState, UnknownContext {

		ContextWrapper localContext;
    	try {
			localContext = contextManager.getContext(context);
		} catch (NoContext e) {
			throw new UnknownContext(e.getMessage());
		} catch (InvalidContext e) {
			throw new InvalidState(e.getMessage());
		}
		RegistrationContext registrationContext = RegistrationContext.augmentToRegistrationContext(localContext.getContext());
		return registrationContext.getParticipants(subProtocols);
	}
	
	public QName getStatus(ContextIdentifier context)
		throws UnknownContext, InvalidState {

		ContextWrapper localContext;
    	try {
			localContext = contextManager.getContext(context, false);
		} catch (NoContext e) {
			throw new UnknownContext(e.getMessage());
		} catch (InvalidContext e) {
			throw new InvalidState(e.getMessage());
		}

		return localContext.getStatus().getState();
	}

	public void registrationReplaced(ContextIdentifier context, EndpointReference registrationService) {
        ContextWrapper localContext;
        try {
            localContext = contextManager.getContext(context);
        } catch (NoContext e) {
            return;
        } catch (InvalidContext e) {
            return;
        }
        RegistrationContext registrationContext = RegistrationContext.augmentToRegistrationContext(localContext.getContext());
        registrationContext.setRegistrationService(registrationService);

	}
	
	private void checkProtocolsSupported(String[] protocols) throws InvalidProtocol {
		if (protocols == null || protocols.length<=0)
			throw new InvalidProtocol("No protocols specified", null);
        
		for (int i=0; i<protocols.length; i++) {
		    boolean supported = false;
            
            for (int j=0; j<protocolsSupported.length; j++)
                if (protocolsSupported[j].equalsIgnoreCase(protocols[i])) {
                    supported=true;
                    break;
                }
            
            if (!supported)
                throw new InvalidProtocol("Protocol not supported by current Registration Service: " + protocols[i], protocols[i]);
        }
	}
    
    private void addCompletionHandlers(ContextWrapper context, String[] protocols) {
        for (int i=0; i<protocols.length; i++)
            context.addCompletionHandler(CompletionHandler.createHandler(protocols[i]));
    }
	
}
