/**
  * JASS Project. WS-CAF.
  *
  * Copyright (c) 2006, Distributed Systems Lab, Universidad Politecnica 
  * de Madrid (Spain)
  * All rights reserved.
  * Contact:
  *        web page: http://lsd.ls.fi.upm.es/lsd
  *        e-mail: mpatino@fi.upm.es or rjimenez@fi.upm.es
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met:
  *
  *    - Redistributions of source code must retain the above copyright 
  *      notice, this list of conditions and the following disclaimer.
  *    - Redistributions in binary form must reproduce the above copyright 
  *      notice, this list of conditions and the following disclaimer in the 
  *      documentation and/or other materials provided with the distribution.
  *    - Neither the name of the Distributed Systems Lab nor the names of 
  *      its contributors may be used to endorse or promote products derived from 
  *      this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, *
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * 
  * Contributors: 
  * Francisco Pérez-Sorrosal (fpsorrosal@zipi.fi.upm.es)
  * Rubén Naranjo Izquierdo (rnaranjo@alumnos.upm.es)
  * José Antonio Sánchez Murillo (jsanchez@alumnos.upm.es)
  * 
  **/

package org.oasis.wscaf.core.wsctx.services;

import java.util.Calendar;

import javax.xml.namespace.QName;

import org.apache.axis2.AxisFault;
import org.oasis.wscaf.core.wsctx.exceptions.ChildActivityPending;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidContext;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidContextStructure;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidProtocol;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidState;
import org.oasis.wscaf.core.wsctx.exceptions.NoContext;
import org.oasis.wscaf.core.wsctx.exceptions.NoPermission;
import org.oasis.wscaf.core.wsctx.exceptions.ParentActivityCompleted;
import org.oasis.wscaf.core.wsctx.exceptions.TimeoutNotSupported;
import org.oasis.wscaf.core.wsctx.types.Context;
import org.oasis.wscaf.core.wsctx.types.ContextIdentifier;
import org.oasis.wscaf.core.wsctx.types.ContextType;
import org.oasis.wscaf.core.wsctx.types.ContextWrapper;
import org.oasis.wscaf.core.wsctx.types.Duration;
import org.oasis.wscaf.core.wsctx.utils.ActivityIdGenerator;

public class ContextService {
	
	// The singleton instance
    private static ContextService singleton = null;

	// Default TimeOut.
	private Duration timeout = new Duration();

	// The generator of activity ids
	private ActivityIdGenerator idGenerator = null;

	// The context manager
    private ContextManager contextManager = null;

	
    /**
     * Private constructor.
     */
    private ContextService() {
        // Get the id generator
        idGenerator = ActivityIdGenerator.getInstance();
        //System.out.println("\tId Generator instance obtained!!!");
        // Get the context manager
        contextManager = ContextManager.getInstance();
        //System.out.println("\t[ContextService] WSCAF CTX Manager instance obtained!!!");
    }

    /**
     * Returns the singleton instance of the context service.
     * 
     * @return The singleton instance.
     */
    public static synchronized ContextService getInstance() {
    	//System.out.println("[ContextService] Returning instance");
        if (singleton == null)
            singleton = new ContextService();
        return singleton;
    }

	public Context begin(Calendar expiresAt, String id, ContextType type) 
		throws InvalidState, InvalidContext, InvalidContextStructure,
			   TimeoutNotSupported, ParentActivityCompleted, NoPermission, 
			   InvalidProtocol {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		return begin(expiresAt, id, null, type);
	}

	public Context begin(Calendar expiresAt, String id, ContextIdentifier parent, ContextType type)
		throws InvalidState, InvalidContext, InvalidContextStructure, 
			   TimeoutNotSupported, ParentActivityCompleted, NoPermission, 
			   InvalidProtocol {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		
		//Protocol type check
		if (type==null || !type.isSupported())
			throw new InvalidProtocol("Protocol type not supported: " + type);
		
		//Timeout check
		boolean defaultExpiresAt = checkCalendar(expiresAt);

		//Parent check
		boolean noParent = false;
    	try {
			contextManager.getContext(parent);
		} catch (NoContext e) {
			noParent = true;
		}

		
		Context context = null;
        ContextIdentifier contextId = idGenerator.newActivityId().toContextIdentifier();

        if (noParent) { // Begin a new root activity
            System.out.println("[ContextService] -------------- BEGIN CONTEXT (ROOT) -------------");

            System.out.println("[ContextService] 1) Creating basic context...");
            if (defaultExpiresAt)
            	context = new Context(contextId, null, getDefaultExpiry(), id, type);
            else
            	context = new Context(contextId, null, expiresAt, id, type);
            
            System.out.println("[ContextService] 2) Adding context to the context manager...");
            contextManager.addContext(context);
        } 
        else {
        	System.out.println("[ContextService] ------------ BEGIN CONTEXT (CHILD) ----------");

        	System.out.println("[ContextService] 1) Creating basic context...");
            if (defaultExpiresAt)
            	context = new Context(contextId, parent, getDefaultExpiry(), id, type);
            else
            	context = new Context(contextId, parent, expiresAt, id, type);
        	
            System.out.println("[ContextService] 2) Adding context to the context manager...");
        	contextManager.addContext(context);
        }

        System.out.println("[ContextService] ----------------- END BEGIN CONTEXT -----------------");

        return context;
	}
	
	public Context complete(ContextIdentifier context, String status) 
		throws AxisFault {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		ContextWrapper localContext = contextManager.getContext(context);

        /* Completion handlers must decide if this exception is thrown
		if (!contextManager.completedChildren(localContext))
    		throw new ChildActivityPending("Trying to complete a context with uncompleted children");
        */

		localContext.complete(status);
		return localContext.getContext();
	}
	
	public QName getStatus(ContextIdentifier context) 
		throws InvalidState, InvalidContext, NoPermission, NoContext {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		return contextManager.getContextStatus(context).getState();
	}
	
	public Duration setTimeout(Duration timeout) 
		throws InvalidState, InvalidContext, TimeoutNotSupported, NoPermission {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		if (checkDuration(timeout)) {
			this.timeout = null;
			return null;
		}
		else {
			this.timeout = timeout;
			return timeout;
		}
	}
	
	public Duration getTimeout() 
		throws InvalidState, InvalidContext, NoPermission {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		return timeout;
	}

    /**
     * Checks if a given calendar is correct or not.
     * 
     * @return true if is the default calendar, false otherwise.
     */
	private boolean checkCalendar (Calendar time) throws TimeoutNotSupported {
		if (time != null) {
			Calendar now = Calendar.getInstance();
			//Comparing with "the epoch"
			if (time.getTime().getTime()==0)
				return true;
			else if (time.before(now))
				throw new TimeoutNotSupported("Date especified is previous to current time", (int)time.getTimeInMillis());
		}
		return false;
	}
	
    /**
     * Checks if a given duration is correct or not.
     * 
     * @return true if is the default duration, false otherwise.
     */
	private boolean checkDuration (Duration time) throws TimeoutNotSupported {
		if (time != null) {
			//Comparing with 0
			if (time.isZero())
				return true;
		}
		return false;
	}
	
    /**
     * Gets the default expiry date by adding "timeout" field to current date.
     * 
     * @return Default expiry date.
     */
	private Calendar getDefaultExpiry() {
		Calendar now = Calendar.getInstance();
		now.add(Calendar.MILLISECOND, timeout.getMilliseconds());
		now.add(Calendar.SECOND,	  timeout.getSeconds());
		now.add(Calendar.MINUTE,	  timeout.getMinutes());
		now.add(Calendar.HOUR,		  timeout.getHours());
		now.add(Calendar.DATE,		  timeout.getDays());
		now.add(Calendar.MONTH,		  timeout.getMonths());
		now.add(Calendar.YEAR,		  timeout.getYears());
		return now;
	}
	
}
